import os, sys, time, ast, shutil
import logging
import traceback
from datetime import datetime, timedelta
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import hashlib
import ms_autoqc.DatabaseFunctions as db
import ms_autoqc.AutoQCProcessing as qc

class DataAcquisitionEventHandler(FileSystemEventHandler):

    """
    Event handler that alerts when the data file has completed sample acquisition.
    For more information, see: https://python-watchdog.readthedocs.io/en/stable/
    """

    def __init__(self, observer, path, filenames, extension, instrument_id, run_id, current_sample):

        self.observer = observer
        self.path = path
        self.filenames = filenames
        self.extension = extension
        self.instrument_id = instrument_id
        self.run_id = run_id
        self.current_sample = current_sample


    def on_created(self, event):

        """
        Listens to data acquisition path and starts watching newly-created data files,
        but only if the filename exists in the sequence.

        If the acquisition listener process was restarted, the last sample being acquired
        or monitored will be reprocessed.

        Args:
            event (FileCreatedEvent): Event representing file / directory creation.

        Returns:
            None
        """

        # Remove directory path and file extension from filename
        path = event.src_path.replace("\\", "/")
        filename = path.split("/")[-1].split(".")[0]

        # For restarted jobs: process the sample that was being acquired when the job was interrupted
        if self.current_sample is not None:
            if os.path.exists(self.path + self.current_sample + "." + self.extension):
                self.trigger_pipeline(self.path, self.current_sample, self.extension)
                self.current_sample = None

        # Route data file to pipeline
        if not event.is_directory and filename in self.filenames:
            self.trigger_pipeline(self.path, filename, self.extension)


    def watch_file(self, path, filename, extension, next_sample=None):

        """
        Returns True if MD5 checksums match AND next sample in sequence has begun acquiring,
        effectively determining whether sample acquisition has been completed.

        Checksum matching is checking if the MD5 checksum computed for the file matches the
        MD5 checksum that was written to the database 3 minutes ago.

        If watching the last sample in the sequence, this function will skip checking for the next sample.

        Args:
            path (str): Data acquisition path
            filename (str): Name of sample data file
            extension (str): Data file extension, derived from instrument vendor
            next_sample (str, default None): Next sample in sequence after current sample being watched

        Returns:
            bool: True if data acquisition is deemed complete.
        """

        # Write initial MD5 checksum to database
        md5_checksum = get_md5(path + filename + "." + extension)
        db.update_md5_checksum(self.instrument_id, filename, md5_checksum)

        # Watch file indefinitely
        while os.path.exists(path):

            # Wait 3 minutes
            print("Waiting 3 minutes...")
            time.sleep(180)

            # Compare checksums
            new_md5 = get_md5(path + filename + "." + extension)
            old_md5 = db.get_md5(self.instrument_id, filename)

            # If the MD5 checksum after 3 mins is the same as before, route to pipeline
            if new_md5 == old_md5:
                print("MD5 checksums matched.")

                if next_sample is None:
                    print("Preparing to process file.")
                    time.sleep(180)
                    return True
                else:
                    if os.path.exists(path + next_sample + "." + extension):
                        print("Preparing to process file.")
                        time.sleep(180)
                        return True
                    else:
                        print("Waiting for instrument to start acquiring next sample:", next_sample)
                        db.update_md5_checksum(self.instrument_id, filename, new_md5)
            else:
                print("MD5 checksums do not match.")
                db.update_md5_checksum(self.instrument_id, filename, new_md5)


    def trigger_pipeline(self, path, filename, extension):

        """
        Wrapper function that routes data file to monitoring and processing functions.

        This function is called every time a data file is created in the data acquisition path.
        See watch_file() and process_data_file() for more information.

        At the end of the instrument run, marks job as completed and kills listener process.

        Args:
            path (str): Data acquisition path
            filename (str): Name of sample data file
            extension (str): Data file extension, derived from instrument vendor

        Returns:
            None
        """

        print("Watching file:", filename)

        # Get next sample
        try:
            next_sample = db.get_next_sample(filename, self.instrument_id, self.run_id)
        except:
            next_sample = None

        # Start watching file until sample acquisition is complete
        try:
            sample_acquired = self.watch_file(path, filename, extension, next_sample)
        except Exception as error:
            print("Error while watching file:", error)
            sample_acquired = None

        # Route data file to MS-AutoQC pipeline
        if sample_acquired:
            print("Data acquisition completed for", filename)
            qc.process_data_file(path, filename, extension, self.instrument_id, self.run_id)
            print("Data processing for", filename, "complete.")

        # Check if data file was the last sample in the sequence
        if filename == self.filenames[-1]:
            # If so, stop acquisition listening
            print("Last sample acquired. Instrument run complete.")
            self.observer.stop()

            # Terminate acquisition listener process
            print("Terminating acquisition listener process.")
            terminate_job(self.instrument_id, self.run_id)


def start_listener(path, instrument_id, run_id):

    """
    Initializes acquisition listener process to process data files upon sample acquisition completion.

    If the QC job is for a completed instrument run (i.e. all data files have been previously acquired and exist
    in a directory), then this function simply iterates through the list of filenames and processes each sample.

    If the QC job is for an active instrument run, this function initializes the Watchdog file monitor to capture
    incoming data files and wait for them to finish writing before processing each sample.

    In addition, to handle crashes, restarted jobs, and other events, this function checks for and processes
    all unprocessed samples in active instrument runs. Unprocessed samples are defined as data files that exist but
    do not have QC results in the database.
    
    For more information on the Watchdog package, see: https://python-watchdog.readthedocs.io/en/stable/

    Args:
        path (str): Data acquisition path
        instrument_id (str): Instrument ID
        run_id (str): Instrument run ID (job ID)

    Returns:
        None
    """

    print("Run monitoring initiated for", path)

    # Check if MS-AutoQC job type is active monitoring or bulk QC
    is_completed_run = db.is_completed_run(instrument_id, run_id)

    # Retrieve filenames for samples in run
    filenames = db.get_remaining_samples(instrument_id, run_id)

    # Get data file extension
    extension = db.get_data_file_type(instrument_id)

    # Format acquisition path
    path = path.replace("\\", "/")
    path = path + "/" if path[-1] != "/" else path

    if is_completed_run:

        # Iterate through files and process each one
        for filename in filenames:

            # If file is not in directory, skip it
            full_path = path + filename + "." + extension

            if not os.path.exists(full_path):
                continue

            # Process data file
            qc.process_data_file(path, filename, extension, instrument_id, run_id)
            print("Data processing for", filename, "complete.")

        print("Last sample acquired. QC job complete.")
        terminate_job(instrument_id, run_id)

    else:
        # Get samples that may have been unprocessed due to an error or accidental termination
        missing_samples, current_sample = db.get_unprocessed_samples(instrument_id, run_id)
        print("Current sample:", current_sample)

        # Check for missed samples and process them before starting file monitor
        if len(missing_samples) > 0:

            # Iterate through files and process each one
            for filename in missing_samples:

                # If file is not in directory, skip it
                full_path = path + filename + "." + extension
                if not os.path.exists(full_path):
                    continue

                # Process data file
                qc.process_data_file(path, filename, extension, instrument_id, run_id)
                print("Data processing for", filename, "complete.")

        # Start file monitor and process files as they are created
        logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")

        observer = Observer()
        event_handler = DataAcquisitionEventHandler(observer, path, filenames, extension, instrument_id, run_id, current_sample)
        observer.schedule(event_handler, path, recursive=True)
        observer.start()

        try:
            while observer.is_alive():
                observer.join(1)
        finally:
            observer.stop()
            observer.join()


def terminate_job(instrument_id, run_id):

    """
    Wraps up QC job after the last data file has been routed to the pipeline.

    Performs the following functions:
        1. Marks instrument run as completed
        2. Uploads database to Google Drive (if Google Drive sync is enabled)
        3. Deletes temporary data file directory in /data
        4. Kills acquisition listener process

    Args:
        instrument_id (str):
            Instrument ID
        run_id (str):
            Instrument run ID (job ID)

    Returns:
        None
    """

    # Mark instrument run as completed
    db.mark_run_as_completed(instrument_id, run_id)

    # Sync database on run completion
    if db.sync_is_enabled():
        db.sync_on_run_completion(instrument_id, run_id)

    # Delete temporary data file directory
    db.delete_temp_directory(instrument_id, run_id)

    # Kill acquisition listener
    pid = db.get_pid(instrument_id, run_id)
    qc.kill_subprocess(pid)


def get_md5(file_path):

    """
    Computes MD5 checksum for a given file.

    Args:
        file_path (str): File path

    Returns:
        str: MD5 checksum for the given file.
    """

    hash_md5 = hashlib.md5()

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


if __name__ == "__main__":

    # Start listening to data file directory
    start_listener(path=sys.argv[1], instrument_id=sys.argv[2], run_id=sys.argv[3])


import ssl
import ms_autoqc.DatabaseFunctions as db
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

def send_message(message):

    """
    Posts a message to the Slack channel registered for notifications in Settings > General.

    For details on the Slack API, see: https://slack.dev/python-slack-sdk/

    Args:
        message (str): Message to send to the Slack channel

    Returns:
        Response object on success, or Error object on failure
    """

    # SSL
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE

    # Retrieve Slack bot token and Slack channel
    slack_token = db.get_slack_bot_token()
    client = WebClient(token=slack_token, ssl=ssl_context)

    # Send Slack message
    try:
        slack_channel = db.get_slack_channel()
        response = client.chat_postMessage(channel=slack_channel, text=message)
        return response
    except SlackApiError as error:
        print("Slack API error:", error)
        return error


import os, json, ast, traceback, time
import plotly.express as px
import pandas as pd
import numpy as np
import ms_autoqc.DatabaseFunctions as db

# Bootstrap color dictionary
bootstrap_colors = {
    "blue": "rgb(0, 123, 255)",
    "red": "rgb(220, 53, 69)",
    "green": "rgb(40, 167, 69)",
    "yellow": "rgb(255, 193, 7)",
    "blue-low-opacity": "rgba(0, 123, 255, 0.4)",
    "red-low-opacity": "rgba(220, 53, 69, 0.4)",
    "green-low-opacity": "rgba(40, 167, 69, 0.4)",
    "yellow-low-opacity": "rgba(255, 193, 7, 0.4)"
}

def get_qc_results(instrument_id, run_id, status="Complete", biological_standard=None, biological_standards_only=False, for_benchmark_plot=False):

    """
    Loads and parses QC results (for samples and biological standards) into Plotly graphs.

    This function will return whatever tables it can in a tuple, and fill None for the tables that throw errors in parsing.
    This is so that an error in retrieving one table will not prevent retrieving other tables.

    Depending on whether Google Drive sync is enabled, this function will load data from either CSV files
    (for active instrument runs) or the local instrument database (for completed runs).

    Regardless of whether Google Drive sync is enabled, the instrument computer (on which the run was started) will
    always load data from its local SQLite database.

    Args:
        instrument_id (str):
            Instrument ID
        run_id (str):
            Instrument run ID (Job ID)
        status (str):
            QC job status, either "Active" or "Complete"
        biological_standard (str, default None):
            If specified, returns QC results for given biological standard associated with job
        biological_standards_only (bool, default False):
            If specified, returns QC results for biological standards only
        for_benchmark_plot (bool, default False):
            If specified, returns QC results specifically for biological standard benchmark plot

    Returns:
        tuple: Tuple containing tables of various sample data in JSON "records" format. Order is as follows:
            1. df_rt_pos: Retention times for internal standards in positive mode
            2. df_rt_neg: Retention times for internal standards in negative mode
            3. df_intensity_pos: Intensities for internal standards in positive mode
            4. df_intensity_neg: Intensities for internal standards in negative mode
            5. df_mz_pos: Precursor masses for internal standards in positive mode
            6. df_mz_neg: Precursor masses for internal standards in negative mode
            7. df_sequence: Acquisition sequence table
            8. df_metadata: Sample metadata table
            9. df_bio_rt_pos: Retention times for targeted features in biological standard sample in positive mode
            10. df_bio_rt_neg: Retention times for targeted features in biological standard sample in negative mode
            11. df_bio_intensity_pos: Intensities for targeted features in biological standard sample in positive mode
            12. df_bio_intensity_neg: Intensities for targeted features in biological standard sample in negative mode
            13. df_bio_mz_pos: Precursor masses for targeted features in biological standard sample in positive mode
            14. df_bio_mz_neg: Precursor masses for targeted features in biological standard sample in negative mode
            15. resources: Metadata for instrument run
            16. df_samples: Table containing sample names, polarities, autosampler positions, and QC results
            17. pos_internal_standards: List of positive mode internal standards
            18. neg_internal_standards: List of negative mode internal standards
            19. df_delta_rt_pos: Delta RT's for internal standards in positive mode
            20. df_delta_rt_neg: Delta RT's for internal standards in negative mode
            21. df_in_run_delta_rt_pos: In-run delta RT's for internal standards in positive mode
            22. df_in_run_delta_rt_neg: In-run delta RT's for internal standards in negative mode
            23. df_delta_mz_pos: Delta m/z's for internal standards in positive mode
            24. df_delta_mz_neg: Delta m/z's for internal standards in negative mode
            25. df_warnings_pos: QC warnings for internal standards in positive mode
            26. df_warnings_neg: QC warnings for internal standards in negative mode
            27. df_fails_pos: QC fails for internal standards in positive mode
            28. df_fails_neg: QC fails for internal standards in negative mode
    """

    # Get run information / metadata
    if db.get_device_identity() != instrument_id and db.sync_is_enabled():
        if status == "Complete":
            load_from = "database"
        elif status == "Active":
            load_from = "csv"
    else:
        load_from = "database"

    if load_from == "database":
        df_run = db.get_instrument_run(instrument_id, run_id)
    elif load_from == "csv":
        db.download_qc_results(instrument_id, run_id)
        df_run = db.get_instrument_run_from_csv(instrument_id, run_id)

    chromatography = df_run["chromatography"].values[0]
    df_sequence = df_run["sequence"].values[0]
    df_metadata = df_run["metadata"].values[0]
    completed = df_run["completed"].astype(int).tolist()[0]

    biological_standards = df_run["biological_standards"].values[0]
    if biological_standards is not None:
        biological_standards = ast.literal_eval(biological_standards)

    # Get internal standards in chromatography method
    precursor_mz_dict = db.get_internal_standards_dict(chromatography, "precursor_mz")
    retention_times_dict = db.get_internal_standards_dict(chromatography, "retention_time")

    resources = {
        "instrument": instrument_id,
        "run_id": run_id,
        "status": status,
        "chromatography": chromatography,
        "precursor_mass_dict": precursor_mz_dict,
        "retention_times_dict": retention_times_dict,
        "samples_completed": completed,
        "biological_standards": json.dumps(biological_standards)
    }

    # Parse m/z, RT, and intensity data for biological standards into DataFrames
    if biological_standards is not None:

        if biological_standard is None:
            biological_standard = biological_standards[0]

        try:
            df_bio_mz_pos = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="precursor_mz", polarity="Pos", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading positive (–) mode biological standard precursor m/z data:", error)
            df_bio_mz_pos = None

        try:
            df_bio_rt_pos = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="retention_time", polarity="Pos", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading positive (–) mode biological standard precursor m/z data:", error)
            df_bio_rt_pos = None

        try:
            df_bio_intensity_pos = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="intensity", polarity="Pos", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading positive (–) mode biological standard retention time data:", error)
            df_bio_intensity_pos = None

        try:
            df_bio_mz_neg = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="precursor_mz", polarity="Neg", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading negative (–) mode biological standard precursor m/z data:", error)
            df_bio_mz_neg = None

        try:
            df_bio_rt_neg = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="retention_time", polarity="Neg", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading positive (–) mode biological standard retention time data:", error)
            df_bio_rt_neg = None

        try:
            df_bio_intensity_neg = db.parse_biological_standard_data(instrument_id=instrument_id, run_id=run_id,
                result_type="intensity", polarity="Neg", biological_standard=biological_standard, load_from=load_from)
        except Exception as error:
            print("Error loading negative (–) mode biological standard intensity data:", error)
            df_bio_intensity_neg = None

    else:
        df_bio_mz_pos = None
        df_bio_rt_pos = None
        df_bio_intensity_pos = None
        df_bio_mz_neg = None
        df_bio_rt_neg = None
        df_bio_intensity_neg = None

    if biological_standards_only:
        return df_bio_rt_pos, df_bio_rt_neg, df_bio_intensity_pos, df_bio_intensity_neg, df_bio_mz_pos, df_bio_mz_neg
    elif for_benchmark_plot:
        return df_bio_intensity_pos, df_bio_intensity_neg

    # Parse m/z, RT, and intensity data for internal standards into DataFrames
    try:
        df_mz_pos = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="precursor_mz", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode precursor m/z data:", error)
        df_mz_pos = None

    try:
        df_rt_pos = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="retention_time", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode retention time data:", error)
        df_rt_pos = None

    try:
        df_intensity_pos = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="intensity", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode intensity data:", error)
        df_intensity_pos = None

    try:
        df_mz_neg = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="precursor_mz", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode precursor m/z data:", error)
        df_mz_neg = None

    try:
        df_rt_neg = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="retention_time", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode retention time data:", error)
        df_rt_neg = None

    try:
        df_intensity_neg = db.parse_internal_standard_data(instrument_id=instrument_id,
            run_id=run_id, result_type="intensity", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode intensity data:", error)
        df_intensity_neg = None

    try:
        df_delta_rt_pos = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Delta RT", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode delta RT data:", error)
        df_delta_rt_pos = None

    try:
        df_delta_rt_neg = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Delta RT", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode delta RT data:", error)
        df_delta_rt_neg = None

    try:
        df_in_run_delta_rt_pos = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="In-run delta RT", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode in-run delta RT data:", error)
        df_in_run_delta_rt_pos = None

    try:
        df_in_run_delta_rt_neg = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="In-run delta RT", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode in-run delta RT data:", error)
        df_in_run_delta_rt_neg = None

    try:
        df_delta_mz_pos = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Delta m/z", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode delta m/z data:", error)
        df_delta_mz_pos = None

    try:
        df_delta_mz_neg = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Delta m/z", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode delta m/z data:", error)
        df_delta_mz_neg = None

    try:
        df_warnings_pos = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Warnings", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode QC warnings data:", error)
        df_warnings_pos = None

    try:
        df_warnings_neg = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Warnings", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (–) mode QC warnings data:", error)
        df_warnings_neg = None

    try:
        df_fails_pos = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Fails", polarity="Pos", load_from=load_from)
    except Exception as error:
        print("Error loading positive (+) mode QC fails data:", error)
        df_fails_pos = None

    try:
        df_fails_neg = db.parse_internal_standard_qc_data(instrument_id=instrument_id,
            run_id=run_id, result_type="Fails", polarity="Neg", load_from=load_from)
    except Exception as error:
        print("Error loading negative (+) mode QC fails data:", error)
        df_fails_neg = None

    # Generate DataFrame for sample table
    try:
        if load_from == "database":
            df_samples = db.get_samples_in_run(instrument_id, run_id, "Both")
        elif load_from == "csv":
            df_samples = db.get_samples_from_csv(instrument_id, run_id, "Both")

        df_samples = df_samples[["sample_id", "position", "qc_result", "polarity"]]
        df_samples = df_samples.rename(
            columns={
                "sample_id": "Sample",
                "position": "Position",
                "qc_result": "QC",
                "polarity": "Polarity"})
        df_samples = df_samples.to_json(orient="records")

    except Exception as error:
        print("Error loading samples from database:", error)
        traceback.print_exc()
        df_samples = ""

    # Get internal standards from data
    if df_rt_pos is not None:
        pos_internal_standards = pd.read_json(df_rt_pos, orient="records").columns.tolist()
        pos_internal_standards.remove("Sample")
    else:
        pos_internal_standards = []

    if df_rt_neg is not None:
        neg_internal_standards = pd.read_json(df_rt_neg, orient="records").columns.tolist()
        neg_internal_standards.remove("Sample")
    else:
        neg_internal_standards = []

    return (df_rt_pos, df_rt_neg, df_intensity_pos, df_intensity_neg, df_mz_pos, df_mz_neg, df_sequence, df_metadata,
        df_bio_rt_pos, df_bio_rt_neg, df_bio_intensity_pos, df_bio_intensity_neg, df_bio_mz_pos, df_bio_mz_neg,
        json.dumps(resources), df_samples, json.dumps(pos_internal_standards), json.dumps(neg_internal_standards),
        df_delta_rt_pos, df_delta_rt_neg, df_in_run_delta_rt_pos, df_in_run_delta_rt_neg, df_delta_mz_pos, df_delta_mz_neg,
        df_warnings_pos, df_warnings_neg, df_fails_pos, df_fails_neg)


def generate_sample_metadata_dataframe(sample, df_rt, df_mz, df_intensity, df_delta_rt, df_in_run_delta_rt,
    df_delta_mz, df_warnings, df_fails, df_sequence, df_metadata):

    """
    Aggregates tables of relevant data from the acquisition sequence, metadata file, and QC results for a selected sample.

    Returns two DataFrames by aggregating the following information:
        1. Acquisition sequence and sample metadata information
        2. Internal standard m/z, RT, intensity, delta m/z, delta RT, in-run delta RT, warnings, and fails

    Args:
        sample (str):
            Sample ID
        df_rt (DataFrame):
            Retention times for internal standards (columns) across samples (rows)
        df_mz (DataFrame):
            Precursor masses for internal standards (columns) across samples (rows)
        df_intensity (DataFrame):
            Intensities for internal standards (columns) across samples (rows)
        df_delta_rt (DataFrame):
            Delta RT's from library values for internal standards (columns) across samples (rows)
        df_in_run_delta_rt (DataFrame):
            Delta RT's from in-run values for internal standards (columns) across samples (rows)
        df_delta_mz (DataFrame):
            Delta m/z's from library values for internal standards (columns) across samples (rows)
        df_warnings (DataFrame):
            QC warnings for internal standards (columns) across samples (rows)
        df_fails (DataFrame):
            QC fails for internal standards (columns) across samples (rows)
        df_sequence (DataFrame):
            Acquisition sequence table
        df_metadata (DataFrame):
            Sample metadata table

    Returns:
        Tuple containing two DataFrames, the first storing internal standard data and the second storing sample metadata.
    """

    df_sample_istd = pd.DataFrame()
    df_sample_info = pd.DataFrame()

    # Index the selected sample, then make sure all columns in all dataframes are in the same order
    columns = df_rt.columns.tolist()
    internal_standards = df_rt.columns.tolist()
    internal_standards.remove("Sample")
    df_sample_istd["Internal Standard"] = internal_standards

    # Precursor m/z
    df_mz = df_mz.loc[df_mz["Sample"] == sample][columns]
    df_mz.drop(columns=["Sample"], inplace=True)
    df_sample_istd["m/z"] = df_mz.iloc[0].astype(float).values.tolist()

    # Retention times
    df_rt = df_rt.loc[df_rt["Sample"] == sample][columns]
    df_rt.drop(columns=["Sample"], inplace=True)
    df_sample_istd["RT"] = df_rt.iloc[0].astype(float).round(2).values.tolist()

    # Intensities
    df_intensity = df_intensity.loc[df_intensity["Sample"] == sample][columns]
    df_intensity.drop(columns=["Sample"], inplace=True)
    intensities = df_intensity.iloc[0].fillna(0).values.tolist()
    df_sample_istd["Intensity"] = ["{:.2e}".format(x) for x in intensities]

    # Delta m/z
    df_delta_mz.replace(" ", np.nan, inplace=True)
    df_delta_mz = df_delta_mz.loc[df_delta_mz["Sample"] == sample][columns]
    df_delta_mz.drop(columns=["Sample"], inplace=True)
    df_sample_istd["Delta m/z"] = df_delta_mz.iloc[0].astype(float).round(6).values.tolist()

    # Delta RT
    df_delta_rt.replace(" ", np.nan, inplace=True)
    df_delta_rt = df_delta_rt.loc[df_delta_rt["Sample"] == sample][columns]
    df_delta_rt.drop(columns=["Sample"], inplace=True)
    df_sample_istd["Delta RT"] = df_delta_rt.iloc[0].astype(float).round(3).values.tolist()

    # In-run delta RT
    df_in_run_delta_rt.replace(" ", np.nan, inplace=True)
    df_in_run_delta_rt = df_in_run_delta_rt.loc[df_in_run_delta_rt["Sample"] == sample][columns]
    df_in_run_delta_rt.drop(columns=["Sample"], inplace=True)
    df_sample_istd["In-Run Delta RT"] = df_in_run_delta_rt.iloc[0].astype(float).round(3).values.tolist()

    # Warnings
    df_warnings.replace(" ", np.nan, inplace=True)
    df_warnings = df_warnings.loc[df_warnings["Sample"] == sample][columns]
    df_warnings.drop(columns=["Sample"], inplace=True)
    df_sample_istd["Warnings"] = df_warnings.iloc[0].astype(str).values.tolist()

    # Fails
    df_fails.replace(" ", np.nan, inplace=True)
    df_fails = df_fails.loc[df_fails["Sample"] == sample][columns]
    df_fails.drop(columns=["Sample"], inplace=True)
    df_sample_istd["Fails"] = df_fails.iloc[0].astype(str).values.tolist()

    if len(df_sequence) > 0:
        df_sequence = df_sequence.loc[df_sequence["File Name"].astype(str) == sample]
        df_sample_info["Sample ID"] = df_sequence["L1 Study"].astype(str).values
        df_sample_info["Position"] = df_sequence["Position"].astype(str).values
        df_sample_info["Injection Volume"] = df_sequence["Inj Vol"].astype(str).values + " uL"
        df_sample_info["Instrument Method"] = df_sequence["Instrument Method"].astype(str).values

    if len(df_metadata) > 0:
        df_metadata = df_metadata.loc[df_metadata["Filename"].astype(str) == sample]
        if len(df_metadata) > 0:
            df_sample_info["Species"] = df_metadata["Species"].astype(str).values
            df_sample_info["Matrix"] = df_metadata["Matrix"].astype(str).values
            df_sample_info["Growth-Harvest Conditions"] = df_metadata["Growth-Harvest Conditions"].astype(str).values
            df_sample_info["Treatment"] = df_metadata["Treatment"].astype(str).values

    df_sample_info = df_sample_info.append(df_sample_info.iloc[0])
    df_sample_info.iloc[0] = df_sample_info.columns.tolist()
    df_sample_info = df_sample_info.rename(index={0: "Sample Information"})
    df_sample_info = df_sample_info.transpose()

    return df_sample_istd, df_sample_info


def generate_bio_standard_dataframe(clicked_sample, instrument_id, run_id, df_rt, df_mz, df_intensity):

    """
    Aggregates data for a selected biological standard.

    TODO: More metrics could be added to sample information cards for biological standards here.

    Aggregates and returns 2 DataFrames for a selected sample:
        1. QC result and causes
        2. Targeted metabolite m/z, RT, intensity, delta RT, and percent change

    Args:
        clicked_sample (str):
            Sample ID
        instrument_id (str):
            Instrument ID
        run_id (str):
            Instrument run ID (job ID)
        df_rt (DataFrame):
            Retention times of targeted metabolites in the biological standard
        df_mz (DataFrame):
            Precursor masses of targeted metabolites in the biological standard
        df_intensity:
            Intensities of targeted metabolites in the biological standard

    Returns:
        Tuple containing two DataFrames, the first storing targeted metabolites data and the second storing sample metadata.
    """

    metabolites = df_mz.columns.tolist()
    metabolites.remove("Name")

    df_sample_features = pd.DataFrame()
    df_sample_features["Metabolite name"] = metabolites
    df_sample_features["Precursor m/z"] = df_mz[metabolites].iloc[0].astype(float).values
    df_sample_features["Retention time (min)"] = df_rt[metabolites].iloc[0].astype(float).round(3).values
    intensities = df_intensity[metabolites].iloc[0].fillna(0).astype(float).values.tolist()
    df_sample_features["Intensity"] = ["{:.2e}".format(x) for x in intensities]

    df_sample_info = pd.DataFrame()
    df_sample_info["Sample ID"] = [clicked_sample]
    qc_result = db.get_qc_results(
        instrument_id=instrument_id, sample_list=[clicked_sample], is_bio_standard=True)["qc_result"].values[0]
    df_sample_info["QC Result"] = [qc_result]

    df_sample_info = df_sample_info.append(df_sample_info.iloc[0])
    df_sample_info.iloc[0] = df_sample_info.columns.tolist()
    df_sample_info = df_sample_info.rename(index={0: "Sample Information"})
    df_sample_info = df_sample_info.transpose()

    return df_sample_features, df_sample_info


def load_istd_rt_plot(dataframe, samples, internal_standard, retention_times):

    """
    Returns line plot figure of retention times (for a selected internal standard) across samples.

    Documentation on Plotly line plots: https://plotly.com/python-api-reference/generated/plotly.express.line.html

    Args:
        dataframe (DataFrame):
            Table of retention times for internal standards (columns) across samples (rows)
        samples (list):
            Samples to query from the DataFrame
        internal_standard (str):
            The selected internal standard
        retention_times (dict):
            Dictionary with key-value pairs of type { internal_standard: retention_time }

    Returns:
        plotly.express.line object: Plotly line plot of retention times (for the selected internal standard) across samples.
    """

    df_filtered_by_samples = dataframe.loc[dataframe["Sample"].isin(samples)]
    df_filtered_by_samples[internal_standard] = df_filtered_by_samples[internal_standard].astype(float).round(3)

    y_min = retention_times[internal_standard] - 0.1
    y_max = retention_times[internal_standard] + 0.1

    fig = px.line(df_filtered_by_samples,
        title="Retention Time vs. Samples – " + internal_standard,
        x=samples,
        y=internal_standard,
        height=600,
        markers=True,
        hover_name=samples,
        labels={"variable": "Internal Standard",
              "index": "Sample",
              "value": "Retention Time"},
        log_x=False)
    fig.update_layout(
        transition_duration=500,
        clickmode="event",
        showlegend=False,
        legend_title_text="Internal Standards",
        margin=dict(t=75, b=75, l=0, r=0))
    fig.update_xaxes(showticklabels=False, title="Sample")
    fig.update_yaxes(title="Retention Time (min)", range=[y_min, y_max])
    fig.add_hline(y=retention_times[internal_standard], line_width=2, line_dash="dash")
    fig.update_traces(hovertemplate="Sample: %{x} <br>Retention Time: %{y} min<br>")

    return fig


def load_istd_intensity_plot(dataframe, samples, internal_standard, treatments):

    """
    Returns bar plot figure of peak intensities (for a selected internal standard) across samples.

    Documentation on Plotly bar plots: https://plotly.com/python-api-reference/generated/plotly.express.bar.html

    Args:
        dataframe (DataFrame):
            Table of intensities for internal standards (columns) across samples (rows)
        samples (list):
            Samples to query from the DataFrame
        internal_standard (str):
            The selected internal standard
        treatments (DataFrame):
            DataFrame with sample treatments (from the metadata file) mapped to sample ID's

    Returns:
        plotly.express.bar object: Plotly bar plot of intensities (for the selected internal standard) across samples.
    """

    df_filtered_by_samples = dataframe.loc[dataframe["Sample"].isin(samples)]

    if len(treatments) > 0:
        # Map treatments to sample names
        df_mapped = pd.DataFrame()
        df_mapped["Sample"] = df_filtered_by_samples["Sample"]
        df_mapped["Treatment"] = df_mapped.replace(
            treatments.set_index("Filename")["Treatment"])
        df_filtered_by_samples["Treatment"] = df_mapped["Treatment"].astype(str)
    else:
        df_filtered_by_samples["Treatment"] = " "

    fig = px.bar(df_filtered_by_samples,
        title="Intensity vs. Samples – " + internal_standard,
        x="Sample",
        y=internal_standard,
        text="Sample",
        color="Treatment",
        height=600)
    fig.update_layout(
        showlegend=False,
        transition_duration=500,
        clickmode="event",
        xaxis=dict(rangeslider=dict(visible=True), autorange=True),
        legend=dict(font=dict(size=10)),
        margin=dict(t=75, b=75, l=0, r=0))
    fig.update_xaxes(showticklabels=False, title="Sample")
    fig.update_yaxes(title="Intensity")
    fig.update_traces(textposition="outside", hovertemplate="Sample: %{x}<br>Intensity: %{y:.2e}<br>")

    return fig


def load_istd_delta_mz_plot(dataframe, samples, internal_standard):

    """
    Returns line plot figure of delta m/z (for a selected internal standard) across samples.

    Documentation on Plotly line plots: https://plotly.com/python-api-reference/generated/plotly.express.line.html

    Args:
        dataframe (DataFrame):
            Table of delta m/z's for internal standards (columns) across samples (rows)
        samples (list):
            Samples to query from the DataFrame
        internal_standard (str):
            The selected internal standard

    Returns:
        plotly.express.line object: Plotly line plot of delta m/z (for the selected internal standard) across samples.
    """

    # Get delta m/z results for selected samples
    df_filtered_by_samples = dataframe.loc[dataframe["Sample"].isin(samples)]

    fig = px.line(df_filtered_by_samples,
        title="Delta m/z vs. Samples – " + internal_standard,
        x=samples,
        y=internal_standard,
        height=600,
        markers=True,
        hover_name=samples,
        labels={"variable": "Internal Standard",
              "index": "Sample",
              "value": "Delta m/z"},
        log_x=False)
    fig.update_layout(
        transition_duration=500,
        clickmode="event",
        showlegend=False,
        legend_title_text="Internal Standards",
        margin=dict(t=75, b=75, l=0, r=0))
    fig.update_xaxes(showticklabels=False, title="Sample")
    fig.update_yaxes(title="delta m/z", range=[-0.01, 0.01])
    fig.update_traces(hovertemplate="Sample: %{x} <br>Delta m/z: %{y}<br>")

    return fig


def load_bio_feature_plot(run_id, df_rt, df_mz, df_intensity):

    """
    Returns scatter plot figure of precursor m/z vs. retention time for targeted features in the biological standard.

    To further clarify:
        x-axis: retention times
        y-axis: precursor masses
        colorscale: percent change in intensity for each feature compared to the average intensity across all runs

    Documentation on Plotly scatter plots: https://plotly.com/python-api-reference/generated/plotly.express.scatter.html

    Args:
        run_id (str):
            Run ID to query the biological standard from
        df_rt (DataFrame):
            Table of retention times for targeted features (columns) across instrument runs (rows)
        df_mz (DataFrame):
            Table of precursor masses for targeted features (columns) across instrument runs (rows)
        df_intensity (DataFrame):
            Table of intensities for targeted features (columns) across instrument runs (rows)

    Returns:
        plotly.express.scatter object: m/z - RT scatter plot for targeted metabolites in the biological standard
    """

    # Get metabolites
    metabolites = df_mz.columns.tolist()
    del metabolites[0]

    # Construct new DataFrame
    bio_df = pd.DataFrame()
    bio_df["Metabolite name"] = metabolites
    bio_df["Precursor m/z"] = df_mz.loc[df_mz["Name"] == run_id][metabolites].iloc[0].astype(float).values
    bio_df["Retention time (min)"] =  df_rt.loc[df_rt["Name"] == run_id][metabolites].iloc[0].astype(float).values
    bio_df["Intensity"] =  df_intensity.loc[df_intensity["Name"] == run_id][metabolites].iloc[0].astype(float).values

    # Get percent change of feature intensities (only for runs previous to this one)
    df_intensity = df_intensity.fillna(0)

    try:
        index_of_run = df_intensity.loc[df_intensity["Name"] == run_id].index.tolist()[0]
        df_intensity = df_intensity[0:index_of_run + 1]
    finally:
        feature_intensity_from_study = df_intensity.loc[df_intensity["Name"] == run_id][metabolites].iloc[0].astype(float).values

    if len(df_intensity) > 1:
        average_intensity_in_studies = df_intensity.loc[df_intensity["Name"] != run_id][metabolites].astype(float).mean().values
        bio_df["% Change"] = ((feature_intensity_from_study - average_intensity_in_studies) / average_intensity_in_studies) * 100
        bio_df.replace(np.inf, 100, inplace=True)
        bio_df.replace(-np.inf, -100, inplace=True)
    else:
        bio_df["% Change"] = 0

    # Plot readiness
    bio_df["Retention time (min)"] = bio_df["Retention time (min)"].round(2)
    bio_df["% Change"] = bio_df["% Change"].round(1).fillna(0)

    labels = {"Retention time (min)": "Retention time (min)",
              "Precursor m/z": "Precursor m/z",
              "Intensity": "Intensity",
              "Metabolite name": "Metabolite name"}

    # Colorscale
    diverging_colorscale = ["#1a88ff", "#3395ff", "#4da3ff", "#a186ca", "#e7727d", "#e35d6a", "#e04958"]
    diverging_colorscale.reverse()

    fig = px.scatter(bio_df,
        title="Biological Standard – Targeted Metabolites",
        x="Retention time (min)",
        y="Precursor m/z",
        height=600,
        hover_name="Metabolite name",
        color="% Change",
        color_continuous_scale=diverging_colorscale,
        labels=labels,
        log_x=False,
        range_color=[-100, 100])
    fig.update_layout(
        showlegend=False,
        transition_duration=500,
        clickmode="event",
        margin=dict(t=75, b=75, l=0, r=0))
    fig.update_xaxes(title="Retention time (min)")
    fig.update_yaxes(title="Precursor m/z")
    fig.update_traces(marker={"size": 30})

    return fig


def load_bio_benchmark_plot(dataframe, metabolite_name):

    """
    Returns bar plot figure of intensities for a targeted metabolite in a biological standard across instrument runs.

    Documentation on Plotly bar plots: https://plotly.com/python-api-reference/generated/plotly.express.bar.html

    Args:
        dataframe (DataFrame):
            Table of intensities for targeted metabolites (columns) across instrument runs (rows)
        metabolite_name (str):
            The targeted metabolite to query from the DataFrame

    Returns:
        plotly.express.bar object: Plotly bar plot of intensities (for the selected targeted metabolite) across instrument runs.
    """

    # Get list of runs
    instrument_runs = dataframe["Name"].astype(str).tolist()

    # Get targeted metabolite intensities for each run
    intensities = dataframe[metabolite_name].values.tolist()
    if len(intensities) == 0:
        intensities = [0 for x in instrument_runs]

    # Get intensities in scientific notation for labeling bar plot
    if intensities is not None:
        intensities_text = ["{:.2e}".format(x) for x in intensities]
    else:
        intensities_text = []

    fig = px.bar(
        x=instrument_runs,
        y=intensities,
        text=intensities_text,
        height=600)
    fig.update_layout(
        title="Biological Standard Benchmark",
        showlegend=False,
        transition_duration=500,
        clickmode="event",
        xaxis=dict(rangeslider=dict(visible=True), autorange=True),
        legend=dict(font=dict(size=10)),
        margin=dict(t=75, b=75, l=0, r=0))
    fig.update_xaxes(title="Study")
    fig.update_yaxes(title="Intensity")
    fig.update_traces(textposition="outside",
                      hovertemplate=f"{metabolite_name}" + "<br>Study: %{x} <br>Intensity: %{text}<br>")

    return fig


def get_internal_standard_index(previous, next, max):

    """
    Button functionality for seeking through internal standards.

    Uses n_clicks from the previous and next buttons to generate an index, which is used to index a list of internal
    standards in the populate_istd_rt_plot(), populate_istd_intensity_plot(), and populate_istd_mz_plot() callback
    functions of the DashWebApp module.

    This function relies on the previous button's n_clicks to be reset to None on every click.

    Args:
        previous (int):
            n_clicks for the "previous" button (None, unless previous button is clicked)
        next (int):
            n_clicks for the "next" button
        max (int):
            Number of internal standards (maximum index for list of internal standards)

    Returns:
        Integer index for a list of internal standards.
    """

    if previous is not None:
        if next is None or next == 0:
            return max - 1

    if previous is None:
        if next is None:
            index = 0
        else:
            index = next
    elif previous is not None:
        index = next - previous

    if index < 0 or index >= max:
        index = 0

    return index


from ms_autoqc.DashWebApp import *

def main():

    """
    Opens web browser and starts Flask server for Dash app
    """

    # Opens localhost:8050 in Google Chrome
    if sys.platform == "win32":
        chrome_path = "C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe"
        webbrowser.register("chrome", None, webbrowser.BackgroundBrowser(chrome_path))
        webbrowser.get("chrome").open("http://127.0.0.1:8050/")
    elif sys.platform == "darwin":
        webbrowser.get("chrome").open("http://127.0.0.1:8050/", new=1)

    # Start Dash app on port 8050
    app.run_server(threaded=False, debug=False, port=8050)

if __name__ == "__main__":
    main()


import warnings
warnings.simplefilter(action="ignore", category=FutureWarning)

import os, time, shutil, psutil, traceback
import pandas as pd
import numpy as np
import ms_autoqc.DatabaseFunctions as db
import ms_autoqc.SlackNotifications as slack_bot

pd.options.mode.chained_assignment = None

def sequence_is_valid(filename, contents, vendor="Thermo Fisher"):

    """
    Validates that instrument sequence file contains the correct columns.

    TODO: Add support for other mass spectrometry instrument vendors here.

    Args:
        filename (str):
            Acquisition sequence file name
        contents (io.StringIO):
            Acquisition sequence as in-memory file object
        vendor (str, default "Thermo Fisher"):
            Instrument vendor for parsing sequence

    Returns:
        True if sequence table is valid, otherwise False.
    """

    if ".csv" not in filename:
        return False

    # Select columns from sequence using correct vendor software nomenclature
    if vendor == "Thermo Fisher":

        # Attempt to load sequence file as a pandas DataFrame
        try:
            df_sequence = pd.read_csv(contents, index_col=False)
        except Exception as error:
            print("Sequence file could not be read.")
            traceback.print_exc()
            return False

        df_sequence.columns = df_sequence.iloc[0]
        df_sequence = df_sequence.drop(df_sequence.index[0])

        # Define required columns and columns found in sequence file
        required_columns = ["File Name", "Path", "Instrument Method", "Position", "Inj Vol"]
        sequence_file_columns = df_sequence.columns.tolist()

        # Check that the required columns are present
        for column in required_columns:
            if column not in sequence_file_columns:
                return False

    return True


def metadata_is_valid(filename, contents):

    """
    Validates that sample metadata file contains the required columns.

    Args:
        filename (str):
            Metadata file name
        contents (io.StringIO):
            Metadata file as in-memory file object

    Returns:
        True if metadata table is valid, otherwise False.
    """

    if ".csv" not in filename:
        return False

    # Attempt to load metadata file as a pandas DataFrame
    try:
        df_metadata = pd.read_csv(contents, index_col=False)
    except Exception as error:
        print("Metadata file could not be read.")
        traceback.print_exc()
        return False

    # Define required columns and columns found in metadata file
    required_columns = ["Filename", "Name from collaborator", "Sample Name", "Species",
                        "Matrix", "Growth-Harvest Conditions", "Treatment"]
    metadata_file_columns = df_metadata.columns.tolist()

    # Check that the required columns are present
    for column in required_columns:
        if column not in metadata_file_columns:
            return False

    return True


def chromatography_valid(chromatography):

    """
    Validates that MSP / TXT files for the given chromatography method exist.

    TODO: Per Brian, some labs don't run in both polarities. Will need to make this function flexible for that.

    Args:
        chromatography (str): Chromatography method ID to validate

    Returns:
        True if chromatography method files exist, False if not.
    """

    # Get chromatography method from database
    df_methods = db.get_chromatography_methods()
    df_methods = df_methods.loc[df_methods["method_id"] == chromatography]

    if len(df_methods) == 0:
        return False

    # Check whether the method's MSP / TXT files exist
    pos_msp_file = db.get_msp_file_path(chromatography, "Positive")
    neg_msp_file = db.get_msp_file_path(chromatography, "Negative")

    if not os.path.isfile(pos_msp_file) or not os.path.isfile(neg_msp_file):
        return False

    return True


def biological_standards_valid(chromatography, biological_standards):

    """
    Validates that the given list of biological standards have MSP files.

    Args:
        chromatography (str):
            Chromatography method ID to validate
        biological_standards (list):
            List of biological standards to validate

    Returns:
        True if all MSP / TXT files exist, False if not.
    """

    # Query the provided biological standards from the database
    df = db.get_biological_standards()
    df = df.loc[df["chromatography"] == chromatography]
    df = df.loc[df["name"].isin(biological_standards)]

    # Check whether the method's MSP / TXT files exist, and return False if they don't
    pos_msp_files = df["pos_bio_msp_file"].astype(str).tolist()
    neg_msp_files = df["neg_bio_msp_file"].astype(str).tolist()
    msp_files = pos_msp_files + neg_msp_files

    for file in msp_files:
        file_path = os.path.join(os.getcwd(), "data", "methods", file)
        if not os.path.isfile(file_path):
            return False

    return True


def convert_sequence_to_json(sequence_contents, vendor="Thermo Fisher"):

    """
    Converts sequence table to JSON string for database storage.

    TODO: Convert to "records" orient instead. Much faster to load data using pd.DataFrame(json.loads(json_string))
        instead of pd.read_json(json_string, orient="split").

    Args:
        sequence_contents (io.StringIO):
            Acquisition file as in-memory file object
        vendor (str, default "Thermo Fisher"):
            Instrument vendor for parsing sequence

    Returns:
        JSON string of acquisition sequence DataFrame in "split" format.
    """

    # Select columns from sequence using correct vendor software nomenclature
    if vendor == "Thermo Fisher":
        df_sequence = pd.read_csv(sequence_contents, index_col=False)
        df_sequence.columns = df_sequence.iloc[0]
        df_sequence = df_sequence.drop(df_sequence.index[0])

    # Convert DataFrames to JSON strings
    return df_sequence.to_json(orient="split")


def convert_metadata_to_json(metadata_contents):

    """
    Converts sequence and metadata files to JSON strings for database storage.

    TODO: Convert to "records" orient instead. Much faster to load data using pd.DataFrame(json.loads(json_string))
        instead of pd.read_json(json_string, orient="split").

    Args:
        metadata_contents (io.StringIO): Metadata file as in-memory file object

    Returns:
        JSON string of sample metadata DataFrame in "split" format.
    """

    # Select columns from metadata
    df_metadata = pd.read_csv(metadata_contents, index_col=False)

    # Convert DataFrames to JSON strings
    return df_metadata.to_json(orient="split")


def run_msconvert(path, filename, extension, output_folder):

    """
    Makes a copy of data file and converts it from instrument vendor format to open mzML format.

    This function runs msconvert.exe in a background process. It checks every second for 30 seconds if the
    mzML file was created, and if it hangs, will terminate the msconvert subprocess and return None.

    TODO: As MS-AutoQC has evolved, some arguments for this function have become redundant.
        The output folder is always fixed, so this parameter should be removed.

    Args:
        path (str):
            Data acquisition path (with "/" at the end)
        filename (str):
            Name of sample data file
        extension (str):
            Data file extension, derived from instrument vendor
        output_folder (str):
            Output directory for mzML file – this is always ../data/instrument_id_run_id/data

    Returns:
        File path for mzML file (*.mzml)
    """

    # Remove files in output folder (if any)
    try:
        for file in os.listdir(output_folder):
            os.remove(output_folder + file)
    except Exception as error:
        print(error)
    finally:
        # Copy original data file to output folder
        shutil.copy2(path + filename + "." + extension, output_folder)

    # Get MSConvert.exe
    try:
        msconvert_folder = db.get_msconvert_directory()
        msconvert_exe = '"' + msconvert_folder + '/msconvert.exe" '
    except:
        print("Failed to locate MSConvert.exe!")
        traceback.print_exc()
        return None

    # Run MSConvert in a subprocess
    command = msconvert_exe + output_folder + filename + "." + extension + " -o " + output_folder
    process = psutil.Popen(command)
    pid = process.pid

    # Check every second for 30 seconds if mzML file was created; if process hangs, terminate and return None
    for index in range(31):
        if not subprocess_is_running(pid):
            break
        else:
            if index != 30:
                time.sleep(1)
            else:
                kill_subprocess(pid)
                return None

    # Delete copy of original data file
    data_file_copy = output_folder + filename + "." + extension
    os.remove(data_file_copy)

    # Return mzML file path to indicate success
    return output_folder + filename + ".mzml"


def run_msdial_processing(filename, msdial_path, parameter_file, input_folder, output_folder):

    """
    Processes data file (in mzML format) using the MS-DIAL console app.

    TODO: As MS-AutoQC has evolved, some arguments for this function have become redundant.
        The input and output folders are fixed, so these parameters should be removed.

    Args:
        filename (str):
            Name of sample data file
        msdial_path (str):
            Path for directory storing MSDialConsoleApp.exe
        parameter_file (str):
            Path for parameters.txt file, stored in /methods directory
        input_folder (str):
            Input folder – this is always ../data/instrument_id_run_id/data
        output_folder (str):
            Output folder – this is always ../data/instrument_id_run_id/results

    Returns:
        File path for MS-DIAL result file (*.msdial)
    """

    # Navigate to directory containing MS-DIAL
    home = os.getcwd()
    os.chdir(msdial_path)

    # Run MS-DIAL in a subprocess
    command = "MsdialConsoleApp.exe lcmsdda -i " + input_folder \
              + " -o " + output_folder \
              + " -m " + parameter_file + " -p"
    process = psutil.Popen(command)
    pid = process.pid

    # Check every second for 30 seconds if process was completed; if process hangs, return None
    for index in range(31):
        if not subprocess_is_running(pid):
            break
        else:
            if index != 30:
                time.sleep(1)
            else:
                kill_subprocess(pid)
                os.chdir(home)
                return None

    # Clear data file directory for next sample
    for file in os.listdir(input_folder):
        filepath = os.path.join(input_folder, file)
        try:
            shutil.rmtree(filepath)
        except OSError:
            os.remove(filepath)

    # Return to original working directory
    os.chdir(home)

    # MS-DIAL output filename
    msdial_result = output_folder + "/" + filename.split(".")[0] + ".msdial"

    # Return .msdial file path
    return msdial_result


def peak_list_to_dataframe(sample_peak_list, df_features):

    """
    Filters duplicates and poor annotations from MS-DIAL peak table and creates DataFrame storing
    m/z, RT, and intensity data for each internal standard (or targeted metabolite) in the sample.

    TODO: Describe duplicate handling in more detail in this docstring.

    Args:
        sample_peak_list (str):
            File path for MS-DIAL peak table, an .msdial file located in /data/instrument_id_run_id/results
        df_features (DataFrame):
            An m/z - RT table derived from internal standard (or biological standard) MSP library in database

    Returns:
        DataFrame with m/z, RT, and intensity data for each internal standard / targeted metabolite in the sample.
    """

    # Convert .msdial file into a DataFrame
    df = pd.read_csv(sample_peak_list, sep="\t", engine="python", skip_blank_lines=True)
    df.rename(columns={"Title": "Name"}, inplace=True)

    # Get only the m/z, RT, and intensity columns
    df = df[["Name", "Precursor m/z", "RT (min)", "Height", "MSMS spectrum"]]

    # Query only internal standards (or targeted features for biological standard)
    feature_list = df_features["name"].astype(str).tolist()
    without_ms2_feature_list = ["w/o MS2:" + feature for feature in feature_list]
    df = df.loc[(df["Name"].isin(feature_list)) | (df["Name"].isin(without_ms2_feature_list))]

    # Label annotations with and without MS2
    with_ms2 = df["MSMS spectrum"].notnull()
    without_ms2 = df["MSMS spectrum"].isnull()
    df.loc[with_ms2, "MSMS spectrum"] = "MS2"

    # Handle annotations made without MS2
    if len(df[with_ms2]) > 0:
        df.replace(["w/o MS2:"], "", regex=True, inplace=True)      # Remove "w/o MS2" from annotation name
        ms2_matching = True                                         # Boolean that says MS/MS was used for identification
    else:
        ms2_matching = False

    # Get duplicate annotations in a DataFrame
    df_duplicates = df[df.duplicated(subset=["Name"], keep=False)]

    # Remove duplicates from peak list DataFrame
    df = df[~(df.duplicated(subset=["Name"], keep=False))]

    # Handle duplicate annotations
    if len(df_duplicates) > 0:

        # Get list of annotations that have duplicates in the peak list
        annotations = df_duplicates[~df_duplicates.duplicated(subset=["Name"])]["Name"].tolist()

        # For each unique feature, choose the annotation that best matches library m/z and RT values
        for annotation in annotations:

            # Get all duplicate annotations for that feature
            df_annotation = df_duplicates[df_duplicates["Name"] == annotation]
            df_feature_in_library = df_features.loc[df_features["name"] == annotation]

            # Calculate delta m/z and delta RT
            df_annotation["Delta m/z"] = df_annotation["Precursor m/z"].astype(float) - df_feature_in_library["precursor_mz"].astype(float).values[0]
            df_annotation["Delta RT"] = df_annotation["RT (min)"].astype(float) - df_feature_in_library["retention_time"].astype(float).values[0]

            # Absolute values
            df_annotation["Delta m/z"] = df_annotation["Delta m/z"].abs()
            df_annotation["Delta RT"] = df_annotation["Delta RT"].abs()

            # First, remove duplicates without MS2 (if an MSP with MS2 spectra was used for processing)
            if ms2_matching:
                new_df = df_annotation.loc[df_annotation["MSMS spectrum"].notnull()]

                # If annotations with MS2 remain, use them moving forward
                if len(new_df) > 1:
                    df_annotation = new_df

                # If no annotations with MS2 remain, filter annotations without MS2 by height
                else:
                    # Choose the annotation with the highest intensity
                    if len(df_annotation) > 1:
                        df_annotation = df_annotation.loc[
                            df_annotation["Height"] == df_annotation["Height"].max()]

                    # If there's only one annotation without MS2 left, choose as the "correct" annotation
                    if len(df_annotation) == 1:
                        df = df.append(df_annotation, ignore_index=True)
                        continue

            # Append the annotation with the lowest delta RT and delta m/z to the peak list DataFrame
            df_rectified = df_annotation.loc[
                (df_annotation["Delta m/z"] == df_annotation["Delta m/z"].min()) &
                (df_annotation["Delta RT"] == df_annotation["Delta RT"].min())]

            # If there is no "best" feature with the lowest delta RT and lowest delta m/z, choose the lowest delta RT
            if len(df_rectified) == 0:
                df_rectified = df_annotation.loc[
                    df_annotation["Delta RT"] == df_annotation["Delta RT"].min()]

            # If the RT's are exactly the same, choose the feature between them with the lowest delta m/z
            if len(df_rectified) > 1:
                df_rectified = df_rectified.loc[
                    df_rectified["Delta m/z"] == df_rectified["Delta m/z"].min()]

            # If they both have the same delta m/z, choose the feature between them with the greatest intensity
            if len(df_rectified) > 1:
                df_rectified = df_rectified.loc[
                    df_rectified["Height"] == df_rectified["Height"].max()]

            # If at this point there's still duplicates for some reason, just choose the first one
            if len(df_rectified) > 1:
                df_rectified = df_rectified[:1]

            # Append "correct" annotation to peak list DataFrame
            df = df.append(df_rectified, ignore_index=True)

    # DataFrame readiness before return
    try:
        df.drop(columns=["Delta m/z", "Delta RT"], inplace=True)
    finally:
        df.reset_index(drop=True, inplace=True)
        return df


def qc_sample(instrument_id, run_id, polarity, df_peak_list, df_features, is_bio_standard):

    """
    Performs quality control on sample data based on user-defined criteria in Settings > QC Configurations.

    The following quality control parameters are used to determine QC pass, warning, or fail:
        1. Intensity dropouts cutoff:
            How many internal standards are missing in the sample?
        2. RT shift from library value cutoff:
            How many retention times are shifted from the expected value for the chromatography method?
        3. RT shift from in-run average cutoff:
            How many retention times are shifted from their average RT during the run?
        4. m/z shift from library value cutoff:
            How many precursor masses are shifted from the expected value for the internal standard?

    This function returns a DataFrame with a single record in the following format:

    | Sample     | Delta m/z | Delta RT | In-run delta RT | Warnings  | Fails |
    | ---------- | --------- | -------- | --------------- | --------- | ----- |
    | SAMPLE_001 | 0.000001  | 0.05     | 0.00001         | Delta RT  | None  |

    Confusingly, a sample has an overall QC result, as well as QC warnings and fails for each internal standard.
    This makes it easier to determine what caused the overall QC result.

    See a screenshot of the sample information card in the Overview page of the website for context.

    While the thresholds for QC pass and fail are explicit, allowing the user to determine thresholds for QC warnings
    was deemed too cumbersome. Instead, an overall QC result of "Warning" happens if any of the following are true:
        1. The number of intensity dropouts is 75% or more than the defined cutoff
        2. The QC result is not a "Fail" and 50% or more internal standards have a QC Warning

    For each internal standard, a note is added to the "Warning" or "Fail" column of qc_dataframe based on the user's
    defined criteria in Settings > QC Configurations. If the internal standard is not marked as a "Fail", then
    "Warnings" for individual internal standards could be marked if:
        1. The delta RT is greater than 66% of the "RT shift from library value" cutoff
        2. The In-run delta RT is greater than 80% of the "RT shift from in-run average value" cutoff
        3. The delta m/z is greater than 80% of the "RT shift from library value" cutoff

    TODO: Define and implement quality control for biological standards.

    Args:
        instrument_id (str):
            Instrument ID
        run_id (str):
            Instrument run ID (Job ID)
        polarity (str):
            Polarity, either "Pos or "Neg"
        df_peak_list (DataFrame):
            Filtered peak table, from peak_list_to_dataframe()
        df_features (DataFrame):
            An m/z - RT table derived from internal standard (or biological standard) MSP library in database
        is_bio_standard (bool):
            Whether sample is a biological standard or not

    Returns:
        (DataFrame, str): Tuple containing QC results table and QC result (either "Pass", "Fail", or "Warning").
    """

    # Handles sample QC checks
    if not is_bio_standard:

        # Refactor internal standards DataFrame
        df_features = df_features.rename(
            columns={"name": "Name",
                     "chromatography": "Chromatography",
                     "polarity": "Polarity",
                     "precursor_mz": "Library m/z",
                     "retention_time": "Library RT",
                     "ms2_spectrum": "Library MS2",
                     "inchikey": "Library INCHIKEY"})

        # Get delta RT and delta m/z values for each internal standard
        df_compare = pd.merge(df_features, df_peak_list, on="Name")
        df_compare["Delta RT"] = df_compare["RT (min)"].astype(float) - df_compare["Library RT"].astype(float)
        df_compare["Delta m/z"] = df_compare["Precursor m/z"].astype(float) - df_compare["Library m/z"].astype(float)
        df_peak_list_copy = df_peak_list.copy()

        # Get MS-DIAL RT threshold and filter out annotations without MS2 that are outside threshold
        with_ms2 = df_compare["MSMS spectrum"].notnull()
        without_ms2 = df_compare["MSMS spectrum"].isnull()
        annotations_without_ms2 = df_compare[without_ms2]["Name"].astype(str).tolist()

        if len(df_compare[with_ms2]) > 0:
            rt_threshold = db.get_msdial_configuration_parameters("Default", parameter="post_id_rt_tolerance")
            outside_rt_threshold = df_compare["Delta RT"].abs() > rt_threshold
            annotations_to_drop = df_compare.loc[(without_ms2) & (outside_rt_threshold)]

            df_compare.drop(annotations_to_drop.index, inplace=True)
            annotations_to_drop = annotations_to_drop["Name"].astype(str).tolist()
            annotations_to_drop = df_peak_list_copy[df_peak_list_copy["Name"].isin(annotations_to_drop)]
            df_peak_list_copy.drop(annotations_to_drop.index, inplace=True)

        # Get in-run RT average for each internal standard
        df_compare["In-run RT average"] = np.nan
        df_run_retention_times = db.parse_internal_standard_data(instrument_id, run_id, "retention_time", polarity, "processing", False)

        if df_run_retention_times is not None:
            # Calculate in-run RT average for each internal standard
            for internal_standard in df_run_retention_times.columns:
                if internal_standard == "Sample":
                    continue
                in_run_average = df_run_retention_times[internal_standard].dropna().astype(float).mean()
                df_compare.loc[df_compare["Name"] == internal_standard, "In-run RT average"] = in_run_average

            # Compare each internal standard RT to in-run RT average
            df_compare["In-run delta RT"] = df_compare["RT (min)"].astype(float) - df_compare["In-run RT average"].astype(float)
        else:
            df_compare["In-run delta RT"] = np.nan

        # Prepare final DataFrame
        qc_dataframe = df_compare[["Name", "Delta m/z", "Delta RT", "In-run delta RT"]]

        # Count internal standard intensity dropouts
        qc_dataframe["Intensity dropout"] = 0
        qc_dataframe["Warnings"] = ""
        qc_dataframe["Fails"] = ""
        for feature in df_features["Name"].astype(str).tolist():
            if feature not in df_peak_list_copy["Name"].astype(str).tolist():
                row = {"Name": feature,
                       "Delta m/z": np.nan,
                       "Delta RT": np.nan,
                       "In-run delta RT": np.nan,
                       "Intensity dropout": 1,
                       "Warnings": "",
                       "Fails": ""}
                qc_dataframe = qc_dataframe.append(row, ignore_index=True)

        # Determine pass / fail based on user criteria
        qc_config = db.get_qc_configuration_parameters(instrument_id=instrument_id, run_id=run_id)
        qc_result = "Pass"

        # QC of internal standard intensity dropouts
        if qc_config["intensity_enabled"].values[0] == 1:

            # Mark fails
            qc_dataframe.loc[qc_dataframe["Intensity dropout"].astype(int) == 1, "Fails"] = "Missing"

            # Count intensity dropouts
            intensity_dropouts = qc_dataframe["Intensity dropout"].astype(int).sum()
            intensity_dropouts_cutoff = qc_config["intensity_dropouts_cutoff"].astype(int).tolist()[0]

            # Compare number of intensity dropouts to user-defined cutoff
            if intensity_dropouts >= intensity_dropouts_cutoff:
                qc_result = "Fail"

            if qc_result != "Fail":
                if intensity_dropouts > intensity_dropouts_cutoff / 1.33:
                    qc_result = "Warning"

        # QC of internal standard RT's against library RT's
        if qc_config["library_rt_enabled"].values[0] == 1:

            # Check if delta RT's are outside of user-defined cutoff
            library_rt_shift_cutoff = qc_config["library_rt_shift_cutoff"].astype(float).values[0]

            # Mark fails
            fails = qc_dataframe["Delta RT"].abs() > library_rt_shift_cutoff
            qc_dataframe.loc[fails, "Fails"] = "RT"

            # Mark warnings
            warnings = ((library_rt_shift_cutoff / 1.5) < qc_dataframe["Delta RT"].abs()) & \
                       ((qc_dataframe["Delta RT"].abs()) < library_rt_shift_cutoff)
            qc_dataframe.loc[warnings, "Warnings"] = "RT"

            if len(qc_dataframe.loc[fails]) >= len(qc_dataframe) / 2:
                qc_result = "Fail"
            else:
                if len(qc_dataframe.loc[warnings]) > len(qc_dataframe) / 2 and qc_result != "Fail":
                    qc_result = "Warning"

        # QC of internal standard RT's against in-run RT average
        if qc_config["in_run_rt_enabled"].values[0] == 1 and df_run_retention_times is not None:

            # Check if in-run delta RT's are outside of user-defined cutoff
            in_run_rt_shift_cutoff = qc_config["in_run_rt_shift_cutoff"].astype(float).values[0]

            # Mark fails
            fails = qc_dataframe["In-run delta RT"].abs() > in_run_rt_shift_cutoff
            qc_dataframe.loc[fails, "Fails"] = "In-Run RT"

            # Mark warnings
            warnings = ((in_run_rt_shift_cutoff / 1.25) < qc_dataframe["In-run delta RT"].abs()) & \
                       (qc_dataframe["In-run delta RT"].abs() < in_run_rt_shift_cutoff)
            qc_dataframe.loc[warnings, "Warnings"] = "In-Run RT"

            if len(qc_dataframe.loc[fails]) >= len(qc_dataframe) / 2:
                qc_result = "Fail"
            else:
                if len(qc_dataframe.loc[warnings]) > len(qc_dataframe) / 2 and qc_result != "Fail":
                    qc_result = "Warning"

        # QC of internal standard precursor m/z against library m/z
        if qc_config["library_mz_enabled"].values[0] == 1:

            # Check if delta m/z's are outside of user-defined cutoff
            library_mz_shift_cutoff = qc_config["library_mz_shift_cutoff"].astype(float).values[0]

            # Mark fails
            fails = qc_dataframe["Delta m/z"].abs() > library_mz_shift_cutoff
            qc_dataframe.loc[fails, "Fails"] = "m/z"

            # Mark warnings
            warnings = ((library_mz_shift_cutoff / 1.25) < qc_dataframe["Delta m/z"].abs()) & \
                       (qc_dataframe["Delta m/z"].abs() < library_mz_shift_cutoff)
            qc_dataframe.loc[warnings, "Warnings"] = "m/z"

            if len(qc_dataframe.loc[fails]) >= len(qc_dataframe) / 2:
                qc_result = "Fail"
            else:
                if len(qc_dataframe.loc[warnings]) > len(qc_dataframe) / 2 and qc_result != "Fail":
                    qc_result = "Warning"

        # Mark annotations without MS2
        qc_dataframe.loc[qc_dataframe["Name"].isin(annotations_without_ms2), "Warnings"] = "No MS2"

    # Handles biological standard QC checks
    else:
        qc_dataframe = pd.DataFrame()
        qc_result = "Pass"

    return qc_dataframe, qc_result


def convert_to_dict(sample_id, df_peak_list, qc_dataframe):

    """
    Converts DataFrames to dictionary records, with features as columns and samples as rows,
    which are then cast to string format for database storage.

    See parse_internal_standard_data() in the DatabaseFunctions module for more information.

    Format:

    | Name       | iSTD 1 | iSTD 2 | iSTD 3 | iSTD 4 | ... |
    | ---------- | ------ | ------ | ------ | ------ | ... |
    | SAMPLE_001 | 1.207  | 1.934  | 3.953  | 8.132  | ... |

    Args:
        sample_id (str):
            Sample ID
        df_peak_list (DataFrame):
            Filtered MS-DIAL peak table result
        qc_dataframe (DataFrame):
            Table of various QC results

    Returns:
        (str, str, str, str): Tuple containing dictionary records of m/z, RT, intensity, and
        QC data, respectively, cast as strings for database storage.
    """

    # m/z, RT, intensity
    df_mz = df_peak_list[["Name", "Precursor m/z"]]
    df_rt = df_peak_list[["Name", "RT (min)"]]
    df_intensity = df_peak_list[["Name", "Height"]]

    df_mz = df_mz.rename(columns={"Precursor m/z": sample_id})
    df_rt = df_rt.rename(columns={"RT (min)": sample_id})
    df_intensity = df_intensity.rename(columns={"Height": sample_id})

    df_mz = df_mz.transpose().reset_index()
    df_rt = df_rt.transpose().reset_index()
    df_intensity = df_intensity.transpose().reset_index()

    df_mz.columns = df_mz.iloc[0].astype(str).tolist()
    df_rt.columns = df_rt.iloc[0].astype(str).tolist()
    df_intensity.columns = df_intensity.iloc[0].astype(str).tolist()

    df_mz = df_mz.drop(df_mz.index[0])
    df_rt = df_rt.drop(df_rt.index[0])
    df_intensity = df_intensity.drop(df_intensity.index[0])

    mz_record = df_mz.to_dict(orient="records")[0]
    rt_record = df_rt.to_dict(orient="records")[0]
    intensity_record = df_intensity.to_dict(orient="records")[0]

    # QC results
    if len(qc_dataframe) > 0:
        for column in qc_dataframe.columns:
            if column != "Name":
                qc_dataframe.rename(columns={column: sample_id + " " + column}, inplace=True)
        qc_dataframe = qc_dataframe.transpose().reset_index()
        qc_dataframe.columns = qc_dataframe.iloc[0].astype(str).tolist()
        qc_dataframe = qc_dataframe.drop(qc_dataframe.index[0])
        qc_dataframe = qc_dataframe.fillna(" ")
        qc_record = qc_dataframe.to_dict(orient="records")
    else:
        qc_record = {}

    return str(mz_record), str(rt_record), str(intensity_record), str(qc_record)


def process_data_file(path, filename, extension, instrument_id, run_id):

    """
    Processes data file upon sample acquisition completion.

    For more details, please visit the Documentation page on the website.

    Performs the following functions:
        1. Convert data file to mzML format using MSConvert
        2. Process data file using MS-DIAL and user-defined parameter configuration
        3. Load peak table into DataFrame and filter out poor annotations
        4. Perform quality control checks based on user-defined criteria
        5. Notify user of QC warnings or fails via Slack or email
        6. Write QC results to instrument database
        7. If Google Drive sync is enabled, upload results as CSV files

    Args:
        path (str):
            Data acquisition path
        filename (str):
            Name of sample data file
        extension (str):
            Data file extension, derived from instrument vendor
        instrument_id (str):
            Instrument ID
        run_id (str):
            Instrument run ID (job ID)

    Returns:
        None
    """

    id = instrument_id.replace(" ", "_") + "_" + run_id

    # Create the necessary directories
    data_directory = os.path.join(os.getcwd(), r"data")
    mzml_file_directory = os.path.join(data_directory, id, "data")
    qc_results_directory = os.path.join(data_directory, id, "results")

    for directory in [data_directory, mzml_file_directory, qc_results_directory]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    mzml_file_directory = mzml_file_directory + "/"
    qc_results_directory = qc_results_directory + "/"

    # Retrieve chromatography, samples, and biological standards using run ID
    df_run = db.get_instrument_run(instrument_id, run_id)
    chromatography = df_run["chromatography"].astype(str).values[0]

    df_samples = db.get_samples_in_run(instrument_id, run_id, sample_type="Sample")
    df_biological_standards = db.get_samples_in_run(instrument_id, run_id, sample_type="Biological Standard")

    # Retrieve MS-DIAL parameters, internal standards, and targeted features from database
    if filename in df_biological_standards["sample_id"].astype(str).tolist():

        # Get biological standard type
        biological_standard = df_biological_standards.loc[
            df_biological_standards["sample_id"] == filename]

        # Get polarity
        try:
            polarity = biological_standard["polarity"].astype(str).values[0]
        except Exception as error:
            print("Could not read polarity from database:", error)
            print("Using default positive mode.")
            polarity = "Pos"

        biological_standard = biological_standard["biological_standard"].astype(str).values[0]

        # Get parameters and features for that biological standard type
        msdial_parameters = db.get_parameter_file_path(chromatography, polarity, biological_standard)
        df_features = db.get_targeted_features(biological_standard, chromatography, polarity)
        is_bio_standard = True

    elif filename in df_samples["sample_id"].astype(str).tolist():

        # Get polarity
        try:
            polarity = df_samples.loc[df_samples["sample_id"] == filename]["polarity"].astype(str).values[0]
        except Exception as error:
            print("Could not read polarity from database:", error)
            print("Using default positive mode.")
            polarity = "Positive"

        msdial_parameters = db.get_parameter_file_path(chromatography, polarity)
        df_features = db.get_internal_standards(chromatography, polarity)
        is_bio_standard = False

    else:
        print("Error! Could not retrieve MS-DIAL libraries and parameters.")
        return

    # Get MS-DIAL directory
    msdial_directory = db.get_msdial_directory()

    # Run MSConvert
    try:
        mzml_file = run_msconvert(path, filename, extension, mzml_file_directory)

        # For active instrument runs, give 3 more attempts if MSConvert fails
        if not db.is_completed_run(instrument_id, run_id):
            for attempt in range(3):
                if not os.path.exists(mzml_file):
                    print("MSConvert crashed, trying again in 3 minutes...")
                    time.sleep(180)
                    mzml_file = run_msconvert(path, filename, extension, mzml_file_directory)
                else:
                    break
    except:
        mzml_file = None
        print("Failed to run MSConvert.")
        traceback.print_exc()

    # Run MS-DIAL
    if mzml_file is not None:
        try:
            peak_list = run_msdial_processing(filename, msdial_directory, msdial_parameters,
                str(mzml_file_directory), str(qc_results_directory))
        except:
            peak_list = None
            print("Failed to run MS-DIAL.")
            traceback.print_exc()

    # Send peak list to MS-AutoQC algorithm if valid
    if mzml_file is not None and peak_list is not None:

        # Convert peak list to DataFrame
        try:
            df_peak_list = peak_list_to_dataframe(peak_list, df_features)
        except:
            print("Failed to convert peak list to DataFrame.")
            traceback.print_exc()
            return

        # Execute AutoQC algorithm
        try:
            qc_dataframe, qc_result = qc_sample(instrument_id, run_id, polarity, df_peak_list, df_features, is_bio_standard)
        except:
            print("Failed to execute AutoQC algorithm.")
            traceback.print_exc()
            return

        # Convert m/z, RT, and intensity data to dictionary records in string form
        try:
            mz_record, rt_record, intensity_record, qc_record = convert_to_dict(filename, df_peak_list, qc_dataframe)
        except:
            print("Failed to convert DataFrames to dictionary record format.")
            traceback.print_exc()
            return

        # Delete MS-DIAL result file
        try:
            os.remove(qc_results_directory + filename + ".msdial")
        except Exception as error:
            print("Failed to remove MS-DIAL result file.")
            traceback.print_exc()
            return

    else:
        print("Failed to process", filename)
        mz_record = None
        rt_record = None
        intensity_record = None
        qc_record = None
        qc_result = "Fail"
        peak_list = None

    # Send email and Slack notification (if they are enabled)
    try:
        if qc_result != "Pass":
            alert = "QC " + qc_result + ": " + filename
            if peak_list is None:
                alert = "Failed to process " + filename

            # Send Slack
            if db.slack_notifications_are_enabled():
                slack_bot.send_message(alert)

            # Send email
            if db.email_notifications_are_enabled():
                db.send_email(alert, "Please check on your instrument run accordingly.")

    except:
        print("Failed to send Slack notification.")
        traceback.print_exc()

    try:
        # Write QC results to database and upload to Google Drive
        db.write_qc_results(filename, instrument_id, run_id, mz_record, rt_record, intensity_record, qc_record, qc_result, is_bio_standard)

        # Update sample counters to trigger dashboard update
        db.update_sample_counters_for_run(instrument_id=instrument_id, run_id=run_id, latest_sample=filename)

        # If sync is enabled, upload the QC results to Google Drive
        if db.sync_is_enabled():
            db.upload_qc_results(instrument_id, run_id)

    except:
        print("Failed to write QC results to database.")
        traceback.print_exc()
        return


def subprocess_is_running(pid):

    """
    Returns True if subprocess is still running, and False if not.

    Args:
        pid (int): Subprocess ID

    Returns:
        bool: True if subprocess is still running, False if not
    """

    if pid is None:
        return False

    time.sleep(1)

    try:
        if psutil.Process(pid).status() == "running":
            return True
        else:
            return False
    except Exception as error:
        return False


def kill_subprocess(pid):

    """
    Kills subprocess.

    Args:
        pid (int): Subprocess ID

    Returns:
        None
    """

    try:
        return psutil.Process(pid).kill()
    except Exception as error:
        print("Error killing acquisition listener.")
        traceback.print_exc()